##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

# XML Elements/Attributes

ELEMENT_ARG = "arg"
ELEMENT_AUTHTYPE = "authtype"
ELEMENT_BODY = "body"
ELEMENT_CALDAVTEST = "caldavtest"
ELEMENT_CALLBACK = "callback"
ELEMENT_CLIENTS = "clients"
ELEMENT_CONTENTTYPE = "content-type"
ELEMENT_DATA = "data"
ELEMENT_DESCRIPTION = "description"
ELEMENT_END = "end"
ELEMENT_EXCLUDE_FEATURE = "exclude-feature"
ELEMENT_FEATURES = "features"
ELEMENT_FEATURE = "feature"
ELEMENT_FILEPATH = "filepath"
ELEMENT_GENERATOR = "generator"
ELEMENT_GRABCALPROP = "grabcalproperty"
ELEMENT_GRABCALPARAM = "grabcalparameter"
ELEMENT_GRABCOUNT = "grabcount"
ELEMENT_GRABELEMENT = "grabelement"
ELEMENT_GRABHEADER = "grabheader"
ELEMENT_GRABJSON = "grabjson"
ELEMENT_GRABPROPERTY = "grabproperty"
ELEMENT_GRABURI = "graburi"
ELEMENT_HEADER = "header"
ELEMENT_HOST = "host"
ELEMENT_HOST2 = "host2"
ELEMENT_KEY = "key"
ELEMENT_LOGGING = "logging"
ELEMENT_MAILFROM = "mailfrom"
ELEMENT_MAILTO = "mailto"
ELEMENT_METHOD = "method"
ELEMENT_NAME = "name"
ELEMENT_NONSSLPORT = "nonsslport"
ELEMENT_NONSSLPORT2 = "nonsslport2"
ELEMENT_NOTIFY = "notify"
ELEMENT_PARENT = "parent"
ELEMENT_PAUSE = "pause"
ELEMENT_PERIOD = "period"
ELEMENT_POINTER = "pointer"
ELEMENT_PROPERTY = "property"
ELEMENT_REPEAT = "repeat"
ELEMENT_REQUEST = "request"
ELEMENT_REQUIRE_FEATURE = "require-feature"
ELEMENT_RUNS = "runs"
ELEMENT_RURI = "ruri"
ELEMENT_SERVERINFO = "serverinfo"
ELEMENT_SPREAD = "spread"
ELEMENT_SSLPORT = "sslport"
ELEMENT_SSLPORT2 = "sslport2"
ELEMENT_START = "start"
ELEMENT_SUBJECT = "subject"
ELEMENT_SUBSTITUTE = "substitute"
ELEMENT_SUBSTITUTIONS = "substitutions"
ELEMENT_SUBSTITUTION = "substitution"
ELEMENT_TEST = "test"
ELEMENT_TESTINFO = "testinfo"
ELEMENT_TESTS = "tests"
ELEMENT_TESTSUITE = "test-suite"
ELEMENT_TIMEOUT = "timeout"
ELEMENT_THREADS = "threads"
ELEMENT_UNIX = "unix"
ELEMENT_UNIX2 = "unix2"
ELEMENT_VALUE = "value"
ELEMENT_VARIABLE = "variable"
ELEMENT_VERIFY = "verify"
ELEMENT_WAITCOUNT = "waitcount"
ELEMENT_WAITDELAY = "waitdelay"
ELEMENT_WAITSUCCESS = "waitsuccess"
ELEMENT_WARNINGTIME = "warningtime"

ATTR_HOST2 = "host2"
ATTR_AUTH = "auth"
ATTR_CHANGE_UID = "change-uid"
ATTR_COUNT = "count"
ATTR_DETAILS = "details"
ATTR_ENABLE = "enable"
ATTR_END_DELETE = "end-delete"
ATTR_GENERATE = "generate"
ATTR_IGNORE = "ignore"
ATTR_IGNORE_ALL = "ignore-all"
ATTR_INTERVAL = "interval"
ATTR_ITERATE_DATA = "iterate-data"
ATTR_NAME = "name"
ATTR_ONLY = "only"
ATTR_PRINT_REQUEST = "print-request"
ATTR_PRINT_RESPONSE = "print-response"
ATTR_PSWD = "pswd"
ATTR_QUOTE = "quote"
ATTR_REQUEST_FAILED = "request-failed"
ATTR_STATS = "stats"
ATTR_SUBSTITUTIONS = "substitutions"
ATTR_TIME_EXCEEDED = "time-exceeded"
ATTR_USER = "user"
ATTR_WAIT_FOR_SUCCESS = "wait-for-success"

ATTR_VALUE_NO = "no"
ATTR_VALUE_RANDOM = "random"
ATTR_VALUE_YES = "yes"
